import abc
from .configuration.interfaces import IConfigurationManager
from .facets.interfaces import IJinjaFacet


class ITarget(abc.ABC):
    """Represents a target."""

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:  # noqa: D105, FNE005
        return (
            hasattr(subclass, "conf") and hasattr(subclass, "jinja") or NotImplemented
        )

    @property
    @abc.abstractmethod
    def conf(self) -> IConfigurationManager:
        """Get the configuration manager."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def jinja(self) -> IJinjaFacet:
        """Get the Jinja facet."""
        raise NotImplementedError



import yaml
import importlib
from copy import deepcopy
from typing import Dict, Any, Optional, Callable, Sequence, cast
from .interfaces import (
    IConfigurationManager,
    IConfigurationContainer,
    IConfigurationSection,
    IPluginConfigurationSection,
    EVENT_CONFIGURATION_CHANGED,
    EVENT_CONFIGURATION_CHANGING,
)
from ..events import EventHandler


def default_get_parent_callable() -> Dict[str, Any]:
    """Return the default parent configuration."""
    return {}


class AbstractConfigurationContainer(IConfigurationContainer):
    """An abstract base class for configuration containers."""

    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
        super().__init__()
        self._config: Dict[str, Any] = config or {}

    def get(self, key: str) -> Any | None:
        """Retrieve the configuration value for a given key."""
        return self._config.get(key)

    def show(self, pattern: Optional[Callable[[str, Any], bool]] = None) -> None:
        """Display the current configuration."""
        if pattern is None:
            pattern = lambda k, v: True  # noqa: E731

        matched: Dict[str, Any] = {}
        for key, value in self._config.items():
            if pattern(key, value):
                matched[key] = value

        print(yaml.dump(matched, default_flow_style=False, sort_keys=False))


class ConfigurationSection(AbstractConfigurationContainer, IConfigurationSection):
    """A configuration section."""

    def __init__(
        self, section_name: str, config: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(config)
        self.__section_name = section_name
        self.__get_parent_callable: Callable[[], Dict[str, Any]] = (
            default_get_parent_callable
        )
        self.__ehandler: EventHandler = EventHandler(
            [EVENT_CONFIGURATION_CHANGING, EVENT_CONFIGURATION_CHANGED], False
        )

    @property
    def section_name(self) -> str:
        """Get the name of the configuration section."""
        return self.__section_name

    def build(self) -> Dict[str, Any]:
        """Build the configuration from a dictionary."""
        built: Dict[str, Any] = self.__get_parent_callable()
        built.update(self._config)
        return built

    def upsert(self, key: str, value: Any) -> None:
        """Insert or updates the configuration for a given key."""
        if not isinstance(key, str):
            raise TypeError("Key must be a string.")

        self.__ehandler.emit(EVENT_CONFIGURATION_CHANGING, key, value)
        self._config[key] = value
        self.__ehandler.emit(EVENT_CONFIGURATION_CHANGED, deepcopy(self._config))

    def patch(self, values: Dict[str, Any]) -> None:
        """Patches the configuration with the provided values."""
        if not isinstance(values, dict):
            raise TypeError("Values must be a dictionary.")

        for key, value in values.items():
            self.upsert(key, value)

        self.__ehandler.emit(EVENT_CONFIGURATION_CHANGED, deepcopy(self._config))

    def show(self, pattern: Optional[Callable[[str, Any], bool]] = None) -> None:
        """Display the current configuration."""
        print(f"Configuration for section '{self.section_name}':")
        super().show(pattern)

    def set_parent_callable(self, parent_func: Callable[[], Dict[str, Any]]) -> None:
        """Set the parent callable that provides additional configuration."""
        if parent_func is None or not callable(parent_func):
            raise TypeError("Parent callable must be a callable function.")

        self.__get_parent_callable = parent_func

    def on_configuration_changing(  # noqa: FNE005
        self, func: Callable[[str, Any], None]
    ) -> bool:
        """Register a function to be called when the configuration is changing."""
        if not callable(func):
            raise TypeError("Function must be callable.")

        return self.__ehandler.link(func, EVENT_CONFIGURATION_CHANGING)

    def on_configuration_changed(  # noqa: FNE005
        self, func: Callable[[Dict[str, Any]], None]
    ) -> bool:
        """Register a function to be called when the configuration has changed."""
        if not callable(func):
            raise TypeError("Function must be callable.")

        return self.__ehandler.link(func, EVENT_CONFIGURATION_CHANGED)


class StringConfigurationSection(ConfigurationSection):
    """A configuration section that only accepts string values."""

    def __init__(
        self, section_name: str, config: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(section_name, config)
        self.on_configuration_changing(lambda _, value: self.__validate_value(value))

    def __validate_value(self, value: Any) -> None:
        """Validate that the value is a string."""
        if not isinstance(value, str):
            raise ValueError("Configuration values must be strings.")


class PluginConfigurationSection(ConfigurationSection, IPluginConfigurationSection):
    """A configuration section specifically for plugins."""

    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
        super().__init__("plugins", config)
        self.on_configuration_changing(lambda _, value: self.__validate_value(value))

    def activate(self, key: str, *args, **kwargs) -> None:
        """Activates the plugin configuration section."""
        module_class = self.get(key)
        if not isinstance(module_class, str):
            raise TypeError(
                f"Plugin configuration for '{key}' must be a string in the format 'module_name, class_name'."
            )

        split = module_class.rsplit(",", 1)
        module_name = split[0].strip()
        class_name = split[1].strip()
        return self.__get_instance(class_name, module_name, *args, **kwargs)

    def __get_instance(
        self,
        class_name: str,
        module_name: str,
        *args,
        **kwargs,
    ) -> Any:
        """Get an instance of a class."""
        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return class_(*args, **kwargs)

    def __validate_value(self, value: Any) -> None:
        """Validate the value for the plugin configuration."""
        if value is None or isinstance(value, str) is False:
            raise ValueError(
                "Plugin configuration values must be strings in the format 'module_name, class_name'."
            )

        split = value.rsplit(",", 1)
        if len(split) != 2 or not all(part.strip() for part in split):
            raise ValueError(
                "Plugin configuration values must be in the format 'module_name, class_name'."
            )


class ConfigurationManager(IConfigurationManager):
    """A configuration manager that manages multiple configuration sections."""

    def __init__(
        self, sections: Optional[Sequence[IConfigurationSection]] = None
    ) -> None:
        super().__init__()
        self.__sections: Dict[str, IConfigurationSection] = {}
        self.__load_sections(sections or [])

    @property
    def plugins(self) -> IPluginConfigurationSection:
        """Retrievesthe plugins configuration section."""
        return cast(IPluginConfigurationSection, self.get_section("plugins"))

    @property
    def value_store(self) -> IConfigurationSection:  # noqa: FNE002
        """Retrieve the value store configuration section."""
        return self.get_section("value_store")

    @property
    def jinja_parameters(self) -> IConfigurationSection:
        """Retrieve the Jinja parameters configuration section."""
        return self.get_section("jinja_parameters")

    @property
    def jinja_macros(self) -> IConfigurationSection:
        """Retrieve the Jinja macros configuration section."""
        return self.get_section("jinja_macros")

    def get_section(self, section_name: str) -> IConfigurationSection:
        """Retrieve a configuration section by name."""
        if section_name not in self.__sections:
            raise KeyError(f"Section '{section_name}' not found.")

        return self.__sections[section_name]

    def register_new_section(
        self, section_name: str, config: Optional[Dict[str, Any]] = None
    ) -> None:
        """Add a new configuration section."""
        if section_name in self.__sections:
            raise ValueError(f"Section '{section_name}' already exists.")

        section = ConfigurationSection(section_name, config)
        self.register_section(section)

    def register_section(self, section: IConfigurationSection) -> None:
        """Register a configuration section."""
        if not isinstance(section, IConfigurationSection):
            raise TypeError("Section must be an instance of IConfigurationSection.")

        if section.section_name in self.__sections:
            raise ValueError(f"Section '{section.section_name}' already exists.")

        self.__sections[section.section_name] = section

    def clear(self) -> None:
        """Clear all configuration sections."""
        self.__sections.clear()
        self.__load_sections([])

    def __load_sections(self, sections: Sequence[IConfigurationSection]) -> None:
        """Load multiple configuration sections."""
        for section in sections:
            if not isinstance(section, IConfigurationSection):
                raise TypeError(
                    "All sections must be instances of IConfigurationSection."
                )

            if (
                section.section_name == "plugins"
                and isinstance(section, IPluginConfigurationSection) is False
            ):

                raise TypeError(
                    "The 'plugins' section must be an instance of IPluginConfigurationSection."
                )

            self.__sections[section.section_name] = section

        if "plugins" not in self.__sections:
            self.__sections["plugins"] = PluginConfigurationSection()

        if "value_store" not in self.__sections:
            self.__sections["value_store"] = ConfigurationSection("value_store")

        if "jinja_parameters" not in self.__sections:
            self.__sections["jinja_parameters"] = ConfigurationSection(
                "jinja_parameters"
            )

        if "jinja_macros" not in self.__sections:
            self.__sections["jinja_macros"] = ConfigurationSection("jinja_macros")



from jinja2 import Environment
from ..configuration.interfaces import IConfigurationContainer
from .interfaces import IJinjaFacet


class JinjaFacet(IJinjaFacet):
    """Implementation of the Jinja facet."""

    def __init__(
        self,
        jinja_parameters: IConfigurationContainer,
        jinja_macros: IConfigurationContainer,
    ) -> None:
        super().__init__()
        self.__parameters: IConfigurationContainer = jinja_parameters
        self.__macros: IConfigurationContainer = jinja_macros

    @property
    def parameters(self) -> IConfigurationContainer:
        """Get the Jinja parameters."""
        return self.__parameters

    @property
    def macros(self) -> IConfigurationContainer:
        """Get the Jinja macros."""
        return self.__macros

    def render(self, template: str, **kwargs) -> str:
        """Render the Jinja template with the given context."""
        env = self.__create_environment()
        render_template = env.from_string(template)
        params = {
            **self.parameters.build(),
            **kwargs,
        }
        return render_template.render(**params)

    def render_file(self, file_path: str, **kwargs) -> str:
        """Render the Jinja template from a file."""
        with open(file_path, "r", encoding="utf-8") as f:
            template = f.read()

        return self.render(template, **kwargs)

    def __create_environment(self) -> Environment:
        """Create the Jinja environment."""
        env = Environment()
        macro_module = env.from_string(self.__combine_macro_values()).module
        for name in dir(macro_module):
            if name.startswith("_"):
                continue

            attr = getattr(macro_module, name)
            if callable(attr):
                env.globals[name] = attr

        return env

    def __combine_macro_values(self) -> str:
        """Combine the macro to the source."""
        # Implementation would go here
        return "\n\n".join(macro for macro in list(self.macros.build().values()))



import types
from typing import Callable, List, Dict, Any


class EventHandler:
    """
    Represents a simple event handler class for synchronous event implementations based registered callback functions.

    Based on (MIT):
        https://github.com/davidvicenteranz/eventhandler/blob/master/eventhandler/__init__.py
    """

    def __init__(self, event_names: list[str], swallow_exceptions: bool = True) -> None:
        self.__events: Dict[str, List[Callable]] = {}
        self.__swallow_exceptions: bool = swallow_exceptions
        if event_names:
            for event in event_names:
                self.register(str(event))

    @property
    def events(self) -> Dict[str, List[Callable]]:
        """Return events as a dictionary."""
        return self.__events

    @property
    def event_list(self) -> List[str]:  # noqa: FNE002
        """Return list of registered events."""
        return list(self.__events.keys())

    @property
    def events_count(self) -> int:  # noqa: FNE002
        """Return number of registered events."""
        return len(self.event_list)

    @staticmethod
    def is_callable(func: Any) -> bool:
        """
        Return true if func is a callable/functions.

        Args:
            func (any): The variable to test if it's a callable.

        Returns:
            bool: A value indicating whether or not func is a callable.
        """
        return isinstance(
            func,
            (
                types.FunctionType,
                types.BuiltinFunctionType,
                types.MethodType,
                types.BuiltinMethodType,
            ),
        )

    def clear_events(self) -> bool:  # noqa: FNE005
        """Clear all events."""
        self.__events = {}
        return True

    def register(self, event_name: str) -> bool:  # noqa: FNE005
        """
        Register an event to which callbacks can be associated.

        Args:
            event_name (str): The name of the event.

        Returns:
            bool: A value indicating if the event was registered successfully.
        """
        if event_name in self.__events:
            return False

        self.__events[event_name] = []
        return True

    def register_link(  # noqa: FNE005
        self, event_name: str, callback: Callable
    ) -> bool:
        """
        Register an event with a callback.

        Args:
            event_name (str): The event name.
            callback (callable): The function to link.

        Returns:
            bool: A value indicating whether or not the registration was successful.
        """
        if not self.is_event_registered(event_name) and not self.register(event_name):
            return False

        return self.link(callback, event_name)

    def unregister(self, event_name: str) -> bool:  # noqa: FNE005
        """
        Remove an event registration.

        Args:
            event_name (str): The event name.

        Returns:
            bool: A value indicating whether or not the event was removed from the registrations.
        """
        if event_name in self.__events:
            del self.__events[event_name]
            return True

        return False

    def is_event_registered(self, event_name: str) -> bool:
        """
        Return if an event is current registered.

        Args:
            event_name (str): The event you want to consult.
        """
        return event_name in self.__events

    def is_callback_in_event(self, event_name: str, callback: Callable) -> bool:
        """
        Return if a given callback is already registered on the events dict.

        Args:
            event_name (str): The event name to look up for the callback inside.
            callback (callable): The callback function to check.

        Returns:
            bool: A value indicating whether or not a given callback is already registered on the events dict.
        """
        return callback in self.__events[event_name]

    def link(self, callback: Callable, event_name: str) -> bool:  # noqa: FNE005
        """
        Link a callback to be executed on fired event..

        Args:
            callback (callable): function to link.
            event_name (str): The event that will trigger the callback execution.

        Returns:
            bool: A value indicating whether or not a given callback was linked.
        """
        if not self.is_callable(callback):
            return False

        if not self.is_event_registered(event_name):
            return False

        if callback not in self.__events[event_name]:
            self.__events[event_name].append(callback)
            return True

        return False

    def unlink(self, callback: Callable, event_name: str) -> bool:  # noqa: FNE005
        """
        Unlink a callback execution from a specific event.

        Args:
            callback (callable): function to link.
            event_name (str): The event that will trigger the callback execution.

        Returns:
            bool: A value indicating whether or not the callback was unlinked
            from the event.
        """
        if not self.is_event_registered(event_name):
            return False

        if callback in self.__events[event_name]:
            self.__events[event_name].remove(callback)
            return True

        return False

    def emit(self, event_name: str, *args, **kwargs) -> bool:  # noqa: FNE005
        """
        Triggers all callback executions linked to the given event.

        Args:
            event_name (str): Event to trigger.
            *args: Arguments to be passed to callback functions execution.
            *kwargs: Keyword arguments to be passed to callback functions execution.

        Returns:
            bool: A value indicating if the event was emitted.
        """
        success = True

        for callback in self.__events[event_name]:
            try:
                callable(callback(*args, **kwargs))
            except Exception as e:
                if not self.__swallow_exceptions:
                    raise e

                success = False
                continue

        return success



import abc
from ..configuration.interfaces import IConfigurationContainer


class IJinjaFacet(metaclass=abc.ABCMeta):
    """Represents a Jinja facet."""

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:  # noqa: D105, FNE005
        return (
            hasattr(subclass, "parameters")
            and hasattr(subclass, "macros")
            and hasattr(subclass, "render")
            and callable(subclass.render)
            and hasattr(subclass, "render_file")
            and callable(subclass.render_file)
            or NotImplemented
        )

    @property
    @abc.abstractmethod
    def parameters(self) -> IConfigurationContainer:
        """Get the Jinja environment."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def macros(self) -> IConfigurationContainer:
        """Get the Jinja macros."""
        raise NotImplementedError

    @abc.abstractmethod
    def render(self, template: str, **kwargs) -> str:
        """Render the Jinja template with the given context."""
        raise NotImplementedError

    @abc.abstractmethod
    def render_file(self, file_path: str, **kwargs) -> str:
        """Render the Jinja template fram a file."""
        raise NotImplementedError



import abc
from typing import Callable, Dict, Any, Optional

EVENT_CONFIGURATION_CHANGING = "configuration_changing"
EVENT_CONFIGURATION_CHANGED = "configuration_changed"


class IConfigurationContainer(metaclass=abc.ABCMeta):
    """Represents a container for configuration sections."""

    @classmethod
    def __subclasshook__(cls, subclass) -> bool:  # noqa: D105, FNE005
        return (
            super().__subclasshook__(subclass)
            and hasattr(subclass, "build")
            and callable(subclass.build)
            and hasattr(subclass, "get")
            and callable(subclass.get)
            and hasattr(subclass, "upsert")
            and callable(subclass.upsert)
            and hasattr(subclass, "patch")
            and callable(subclass.patch)
            and hasattr(subclass, "show")
            and callable(subclass.show)
            or NotImplemented
        )

    @abc.abstractmethod
    def build(self) -> Dict[str, Any]:
        """Build the configuration from a dictionary."""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, key: str) -> Any | None:
        """Retrieve the configuration for a given key."""
        raise NotImplementedError

    @abc.abstractmethod
    def upsert(self, key: str, value: Any) -> None:
        """Insert or updates the configuration for a given key."""
        raise NotImplementedError

    @abc.abstractmethod
    def patch(self, values: Dict[str, Any]) -> None:
        """Patche the configuration for a given key."""
        raise NotImplementedError

    @abc.abstractmethod
    def show(self, pattern: Optional[Callable[[str, Any], bool]] = None) -> None:
        """Display the current configuration."""
        raise NotImplementedError


class IConfigurationSection(IConfigurationContainer):
    """Represents a section of configuration."""

    @classmethod
    def __subclasshook__(cls, subclass) -> bool:  # noqa: D105, FNE005
        return (
            super().__subclasshook__(subclass)
            and hasattr(subclass, "section_name")
            and hasattr(subclass, "set_parent_callable")
            and callable(subclass.set_parent_callable)
            and hasattr(subclass, "on_configuration_changing")
            and callable(subclass.on_configuration_changing)
            and hasattr(subclass, "on_configuration_changed")
            and callable(subclass.on_configuration_changed)
            or NotImplemented
        )

    @property
    @abc.abstractmethod
    def section_name(self) -> str:
        """Get the name of the configuration section."""
        raise NotImplementedError

    @abc.abstractmethod
    def set_parent_callable(self, parent_func: Callable[[], Dict[str, Any]]) -> None:
        """Set a function to get the parent configuration."""
        raise NotImplementedError

    @abc.abstractmethod
    def on_configuration_changing(  # noqa: FNE005
        self, func: Callable[[str, Any], None]
    ) -> bool:
        """Call when the configuration is about to change."""
        raise NotImplementedError

    @abc.abstractmethod
    def on_configuration_changed(  # noqa: FNE005
        self, func: Callable[[Dict[str, Any]], None]
    ) -> bool:
        """Call when the configuration has changed."""
        raise NotImplementedError


class IPluginConfigurationSection(IConfigurationSection):
    """Represents a plugin configuration section."""

    @classmethod
    def __subclasshook__(cls, subclass) -> bool:  # noqa: D105, FNE005
        return (
            super().__subclasshook__(subclass)
            and hasattr(subclass, "activate")
            and callable(subclass.activate)
            or NotImplemented
        )

    @abc.abstractmethod
    def activate(self, key: str, *args, **kwargs) -> Any:
        """Activates the configured plugin."""
        raise NotImplementedError


class IConfigurationManager(metaclass=abc.ABCMeta):
    """Represents a configuration manager."""

    @classmethod
    def __subclasshook__(cls, subclass) -> bool:  # noqa: D105, FNE005
        return (
            hasattr(subclass, "plugins")
            and hasattr(subclass, "value_store")
            and hasattr(subclass, "jinja_parameters")
            and hasattr(subclass, "jinja_macros")
            and hasattr(subclass, "get_section")
            and callable(subclass.get_section)
            and hasattr(subclass, "create_new_section")
            and callable(subclass.create_new_section)
            and hasattr(subclass, "register_section")
            and callable(subclass.register_section)
            and hasattr(subclass, "clear")
            and callable(subclass.clear)
            or NotImplemented
        )

    @property
    @abc.abstractmethod
    def plugins(self) -> IPluginConfigurationSection:
        """Return the plugin configuration section."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def value_store(self) -> IConfigurationSection:  # noqa: FNE002
        """Return the value store configuration section."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def jinja_parameters(self) -> IConfigurationSection:
        """Return the Jinja parameters configuration section."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def jinja_macros(self) -> IConfigurationSection:
        """Return the Jinja macros configuration section."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_section(self, section_name: str) -> IConfigurationSection:
        """Retrieve a configuration section by name."""
        raise NotImplementedError

    @abc.abstractmethod
    def register_new_section(
        self, section_name: str, config: Optional[Dict[str, Any]] = None
    ) -> None:
        """Create a new configuration section."""
        raise NotImplementedError

    @abc.abstractmethod
    def register_section(self, section: IConfigurationSection) -> None:
        """Register a new configuration section."""
        raise NotImplementedError
